/**
 * Scraper written in Node.js.
 * Recursively traverses folders and bundles image filepaths into a single output file.
 */

// Arguments used by the program.
var args = require('minimist')(process.argv.slice(2));
// Framework for pretty printing.
var columnify = require('columnify');
// Framework for colorful output.
var colors = require('colors/safe');
// Internal utils for object debugging.
var util = require('util');
// Path util for constructing file paths.
var path = require('path');
// Filesystem util required for working with files.
var fs = require('fs');

// The working directory of the terminal invoking this script.
var exeDir = path.resolve(process.cwd());
// The name of the per-directory configuration file.
var folderConfigFile = "_aGallery.json";
// Every image type to scrape.
var validExtensions = [".gif",
    ".jpg",
    ".jpeg",
    ".png",
    ".webm"];
// Import the SceneMap class from the game.
eval(fs.readFileSync(path.resolve(path.join(exeDir, "res", "js", "SceneMap.js")), 'utf8'));
var sceneMap = new SceneMap();

/**
 * Iterates over the passed array of elements which may have embedded arrays, and applies the function to each element.
 * @param {function} fun The function to apply. Will be passed the current element.
 * @param {array} elems The elements to iterate over.
 */
function recArrayTraverse(fun, elems) {
    for (var i = 0; i < elems.length; i++) {
        var elem = elems[i];
        if (elem && elem.constructor === Array) {
            recArrayTraverse(fun, elem);
        } else {
            fun(elem);
        }
    }
}

/**
 * Logs the passed strings as errors, and then terminates the program.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function error(...strings) {
    recArrayTraverse(elem => console.log(colors.red("ERROR: " + elem)), strings);
    process.exit(1);
}

/**
 * Logs the passed strings as warnings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function warn(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.yellow("WARNING: " + elem)), strings);
}

/**
 * Logs the passed strings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function log(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(elem), strings);
}

/**
 * Logs the passed strings as debug messages.
 * Does nothing if debugging or output in general is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function debug(...strings) {
    if (args.q || !args.debug) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.green("DEBUG: " + elem)), strings);
}

/**
 * Prints the help dialogue.
 */
function printHelp() {
    var argumentHelp = {
        "output": "Output filename. Should always be passed first.",
        "-h": "Displays this help message. Does not run aGalleryScraper.",
        "-q": "Runs aGalleryScraper silently (with no output).",
        "--debug": "Displays debug information while running."
    };
    log("aGalleryScraper is a scraper for Kobold Adventure images.");
    log("It traverses scenes and finds the filenames of all images within.");
    log("This tool MUST be ran from the root of the Kobold Adventure file structure.");
    log(columnify(argumentHelp, {columns: ['FLAG', 'EXPLANATION'], config: {FLAG: {minWidth: 20}, EXPLANATION: {maxWidth: process.stdout.columns - 40}}}));
}

/**
 * Checks if the passed parameter value is a string.
 * If it's not, the program exits with an error.
 * @param {string} toCheck The variable to check.
 * @param {string} flag Which flag is being checked. Produces a more informative error message.
 */
function stringOrExit(toCheck, flag) {
    if (typeof toCheck !== "string") {
        error("", `Parameter ${flag} must be a string value.`, `Example: ${flag} test`, "");
    }
}

/**
 * Gathers and returns all of the initial targets for aGalleryScraper by inspecting the command line parameters.
 * @returns {array} An array of file and folder names to target.
 */
function buildInitialTargets() {
    debug("", "Building initial targets.");
    var output = [];

    var allScenes = sceneMap.getAll();
    for (var i = 0; i < allScenes.length; i++) {
        output.push(allScenes[i]);
    }

    debug("Initial targets:", output, "");
    return output;
}

/**
 * Returns true if the target is an existing file or folder, false otherwise.
 * @param {string} file A path to a file or folder.
 * @returns {boolean} True if the target exists, false otherwise.
 */
function exists(file) {
    return fs.existsSync(file);
}

/**
 * Exits the program with an error message if the passed file does not exist.
 * @param {string} file A path to a file or folder.
 */
function existsOrExit(file) {
    if (!exists(file)) {
        error("", "File not found:", file, "");
    }
}

/**
 * Gets filestats for the specified file or folder.
 * @param {string} file The path to the file or folder to get stats for.
 * @returns {fs.Stats} The file's stats.
 */
function fileStats(file) {
    return fs.lstatSync(file);
}

/**
 * Returns true if the file associated with these stats is indeed a file.
 * @param {fs.Stats} stats The file's stats.
 * @returns {boolean} True if the file is actually a file, false otherwise.
 */
function isFile(stats) {
    return stats.isFile();
}

/**
 * Returns true if the file associated with these stats is a folder.
 * @param {fs.Stats} stats The file's stats.
 * @returns {boolean} True if the file is a folder, false otherwise.
 */
function isFolder(stats) {
    return stats.isDirectory();
}

/**
 * Reads the _aGallery.json file of the specified folder (if there is one) and returns it converted into an object.
 * Does nothing if there is no buildfile. Crashes the program is there is a build file, but it's malformed.
 * @param {string} target A path to a folder.
 * @returns {object} An object describing the aBuild file.
 */
function readBuildFile(target) {
    var aGalleryPath = path.resolve(path.join(target, folderConfigFile));
    if (!exists(aGalleryPath)) {
        return {};
    }
    log("Reading " + aGalleryPath);
    var aGallery = fs.readFileSync(aGalleryPath, 'utf8');
    try {
        aGallery = JSON.parse(aGallery);
    } catch (e) {
        error("", "Failed to parse build file " + aGalleryPath, "Make sure the JSON file is well-structured.", "");
    }
    debug("Parsed: " + util.inspect(aGallery));
    return aGallery;
}

/**
 * Alters the constructed targets-object for the specified target-directory by reading its _aGallery.json file.
 * Does nothing if there is no _aGallery.json file.
 * @param {string} target A path to a folder.
 * @param {object} targets An object containing the reqs and the targets for the specified folder. May be altered by this function.
 */
function folderSpecificBuildFile(target, targets) {
    var aGallery = readBuildFile(target);
    var includes = aGallery.inc && aGallery.inc.constructor === Array ? aGallery.inc : [];
    var excludes = aGallery.exc && aGallery.exc.constructor === Array ? aGallery.exc : [];
    excludes.push(folderConfigFile); // Ensure the aGallery file is not included in the output unless specifically included.
    for (var i = 0; i < includes.length; i++) {
        var filePath = path.resolve(path.join(target, includes[i]));
        targets.reqs.push(filePath);
        excludes.push(filePath); // Make sure it's no longer in targets.
    }
    for (var i = 0; i < excludes.length; i++) {
        var filePath = path.resolve(path.join(target, excludes[i]));
        var index = targets.targets.indexOf(filePath);
        index >= 0 && targets.targets.splice(index, 1);
    }
    if (aGallery.skip) {
        targets.targets = [];
    }
}

/**
 * Builds and returns an object containing the reqs and the targets for the specified folder.
 * @param {string} target A path to a folder.
 * @returns {object} An object containing the reqs and the targets for the specified folder.
 */
function buildTargetsForFolder(target) {
    var output = {reqs: [], targets: []};
    var files = fs.readdirSync(target);
    for (var i = 0; i < files.length; i++) {
        var absoluteFile = path.resolve(path.join(target, files[i]));
        output.targets.push(absoluteFile);
    }
    folderSpecificBuildFile(target, output);
    return output;
}

/**
 * Compiles a single file, if it has not yet been encountered.
 * Prints a warning if the file has already been compiled.
 * @param {string} target A path to the file to compile.
 * @returns {string} 
 */
function compileFile(target) {
    if (compile.visited.includes(target)) {
        warn("File already compiled: " + target);
        return "";
    }
    compile.visited.push(target);
    log("Scraping " + target);
    return target.replace(exeDir, "").substr(1);
}

/**
 * Compiles a single folder, if it has not yet been encountered.
 * Prints a warning if the folder has already been compiled.
 * Compiles by creating a new list of reqs and targets and recursively calling the compile function with these new lists.
 * @param {string} target A path to the folder to compile.
 * @returns {array} 
 */
function compileFolder(target) {
    if (compile.visited.includes(target)) {
        warn("File already Scraped: " + target);
        return "";
    }
    compile.visited.push(target);
    debug("Scraping folder " + target);
    var targets = buildTargetsForFolder(target);
    return compile(targets.reqs, targets.targets);
}

/**
 * Compiles the passed required target, unless it has already been compiled.
 * Returns the compiled string.
 * Prints an error and exits the program if the file or folder can not be found.
 * @param {string} req A single required file or directory.
 * @returns {array} 
 */
function compileReq(req) {
    var output = [];
    existsOrExit(req);
    var stats = fileStats(req);
    if (isFolder(stats)) {
        output = output.concat(compileFolder(req));
    } else if (isFile(stats)) {
        output.push(compileFile(req));
    } else {
        warn("Skipped because neither folder nor file: " + req);
    }
    return output;
}

/**
 * Compiles the passed target, unless it is filtered out by settings or unless it has already been compiled.
 * Returns the compiled string.
 * Prints an error and exits the program if the file or folder can not be found.
 * @param {string} target A single file or directory.
 * @returns {array} 
 */
function compileTarget(target) {
    var output = [];
    existsOrExit(target);
    var stats = fileStats(target);
    if (isFolder(stats)) {
        output = output.concat(compileFolder(target));
    } else if (isFile(stats)) {
        if (validExtensions.includes(path.extname(target))) {
            output.push(compileFile(target));
        }
    } else {
        warn("Skipped because neither folder nor file: " + target);
    }
    return output;
}

/**
 * Recursively compiles all the specified required targets.
 * Returns the compiled string.
 * @param {array} reqs All the required targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files are guaranteed inclusion in the output, regardless of settings.
 * @returns {array} 
 */
function compileReqs(reqs) {
    var output = [];
    for (var i = 0; i < reqs.length; i++) {
        output = output.concat(compileReq(reqs[i]));
    }
    return output;
}

/**
 * Recursively compiles all the specified targets.
 * Returns the compiled string.
 * @param {array} targets All the targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files may be omitted from the output depending on settings.
 * @returns {array} 
 */
function compileTargets(targets) {
    var output = [];
    for (var i = 0; i < targets.length; i++) {
        output = output.concat(compileTarget(targets[i]));
    }
    return output;
}

/**
 * Recursively compiles all the specified targets.
 * Returns the compiled string.
 * initCompile() should always be called before the initial call to this function.
 * @param {array} reqs All the required targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files are guaranteed inclusion in the output, regardless of settings.
 * @param {array} targets All the targets to compile. These may be directories or files. Directories will be recursively scanned.
 *                     Files may be omitted from the output depending on settings.
 * @returns {array} The scraped filenames as an array.
 */
function compile(reqs, targets) {
    var output = [];
    output = output.concat(compileReqs(reqs));
    output = output.concat(compileTargets(targets));
    return output;
}

/**
 * Scrapes a single scene for images.
 * @param {string} scene The name of the scene to scrape.
 * @returns {array}
 */
function scrapeScene(scene) {
    return compile([sceneMap.constructFolderPath(scene)], []);
}

/**
 * Scrapes all of the passed scenes for image files and gathers the image files grouped by their scene names in an object.
 * @param {array} scenes The names of the scenes to scrape.
 * @returns {object} An object containing the paths of all oof the image files grouped by their scene names.
 */
function scrapeScenes(scenes) {
    var output = {};
    for (var i = 0; i < scenes.length; i++) {
        output[scenes[i]] = scrapeScene(scenes[i]);
    }
    return output;
}

/**
 * Initializes things needed for compilation.
 * Should always be called before the initial call to compile.
 */
function initCompile() {
    compile.visited = [];
    log("Compiler initialized.");
}

if (args.h) {
    printHelp();
    process.exit(0);
}

log(colors.cyan.underline("aGalleryScraper"));
debug("", "----------------------", "Debug logging started.", "----------------------", "");
debug("Arguments:", util.inspect(args), "");

initCompile();
console.time("Duration");
log("Commencing compilation.");

var scraped = JSON.stringify(scrapeScenes(buildInitialTargets()), null, 4);
scraped = `// Automatically generated file. Do not alter.
function GalleryContents() {
this.contents =
${scraped};
}`;

log("Compilation complete.");
console.timeEnd("Duration");

if (args._ && args._.length > 0) {
    var outputFileName = path.resolve(args._[0]);
    log("Exporting to " + outputFileName);
    fs.writeFileSync(outputFileName, scraped);
    log("Output file successfully created.");
} else {
    log(scraped);
}

